/* event.c -- Event Manager
   Copyright 2000, 2001, 2002 Free Software Foundation, Inc.
   Written by Stephane Carrez (stcarrez@worldnet.fr)

This file is part of GEL.

GEL is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GEL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GEL; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <sys/param.h>
#include <sys/locks.h>
#include <sys/interrupts.h>
#include <sys/ports.h>
#include <gel/event.h>
#include <gel/timer.h>

static event *event_queue_first PAGE0_ATTRIBUTE;
static event *event_queue_wr PAGE0_ATTRIBUTE;
static event *event_queue_rd PAGE0_ATTRIBUTE;
static event *event_queue_last PAGE0_ATTRIBUTE;

unsigned long event_idle_counter PAGE0_ATTRIBUTE;

unsigned char event_false = 0;

/* Returns a pointer to the next event object in the queue.  */
static inline event *
next_event (event *p)
{
  p++;
  if (p == event_queue_last)
    p = event_queue_first;

  return p;
}

/** Initialize the event package.

    @param queue pointer to the start of the event queue buffer
    @param size  number of elements that can be stored in that buffer
 */
void
event_initialize (event *queue, unsigned short size)
{
  event_queue_first = queue;
  event_queue_wr    = queue;
  event_queue_rd    = queue;
  event_queue_last  = &queue[size];
}

/* Post an event of type 'type'.

   When an event is posted, it is inserted in the event queue.  Events
   are dispatched by <b>event_loop</b>.

   @param type the type of event
   @param data
*/
void
event_post (event_type type, unsigned short data)
{
  unsigned short mask;
  
  /* If no callout is attached to this type of event, drop them.  */
  mask = lock ();
  if (type->callouts != 0)
    {
      event_queue_wr->data = data;
      event_queue_wr->type = type;
      event_queue_wr->time = get_timer_counter ();

      event_queue_wr = next_event (event_queue_wr);
      /* Drop the oldest event.  */
      if (event_queue_wr == event_queue_rd)
        {
          event_queue_rd = next_event (event_queue_rd);
        }
    }
  restore (mask);
}

/** Register an event handler.
    
    Register an handler to receive events of a given type.

    @param type the type of the event to subscribe
    @param callout
    @param handler operation to be called when the event is received
    @param client_data specifies the argument of the handler
*/
void
event_add_callout (event_type type, event_callout *callout)
{
  unsigned short mask;
  
  callout->type = type;
  mask = lock ();
  callout->next = type->callouts;
  type->callouts = callout;
  restore (mask);
}

void
event_remove_callout (event_callout *callout)
{
  unsigned short mask;
  event_callout *prev;
  event_type type;

  type = callout->type;
  mask = lock ();
  prev = type->callouts;
  if (callout == prev)
    {
      type->callouts = prev->next;
    }
  else
    {
      while (prev && prev->next != callout)
        prev = prev->next;

      if (prev)
        prev->next = callout->next;
    }
  restore (mask);
}

static void
event_dispatch (char *bool)
{
  unsigned short mask;

  do
    {
      event ev;
      event_callout *callout;
      
      mask = lock ();
      if (event_queue_rd == event_queue_wr)
        {
          restore (mask);
          break;
        }
      ev = *event_queue_rd;
      event_queue_rd = next_event (event_queue_rd);
      restore (mask);

      /* Assumption: interrupt handlers must not register or
         unregister an event handler.  */
      
      callout = ev.type->callouts;
      while (callout)
        {
          event_callout *next = callout->next;

          callout->callback (callout, &ev);
          callout = next;
        }
    }
  while (*bool == 0);
}

static void
event_notify_received (event_callout *callout, event *ev)
{
  char *bool = (char*) callout->data;

  *bool = 1;
}

/** Dispatches events until a condition becomes true or a delay ellapses.

    This is a main loop of the event module.

    @param nticks the maximum number of ticks to dispatch events
    @param bool   the boolean variable that must be checked
*/
void
event_loop (unsigned long nticks, unsigned char *bool)
{
  unsigned long  timeout_ovf;
  unsigned short timeout_tcnt;
  unsigned short mask;
  unsigned short cur_tcnt;

  /* Get current time expressed in number of timer overflow and
     in current timer counter.  */
  mask = lock ();
  timeout_ovf = timer_current_overflow ();
  cur_tcnt = get_timer_counter ();
  restore (mask);

  /* Compute the timer counter value when the timeout will expire.
     Take into account a possible overflow while we wait.  */
  timeout_tcnt = (unsigned short) (nticks & 0x0ffffL);
  timeout_tcnt += cur_tcnt;
  if (cur_tcnt > timeout_tcnt)
    timeout_ovf++;

  /* Compute the timer overflow value when the timeout will expire.
     Handle infinite timeout by using the infinite overflow value.  */
  timeout_ovf += nticks >> 16;
  if (timeout_ovf >= 0x0ffff)
    timeout_ovf = (unsigned long) -1;
  
  while (1)
    {
      unsigned long cur_ovf;

      event_idle_counter++;
      event_dispatch (bool);

      if (*bool != 0)
        break;

      mask = lock ();
      cur_ovf = timer_current_overflow ();
      cur_tcnt = get_timer_counter ();
      restore (mask);

      if (cur_ovf > timeout_ovf)
	break;

      if (cur_tcnt >= timeout_tcnt && cur_ovf == timeout_ovf)
	break;
    }
}

/** Wait until an event occurs or a delay ellapses.

    @param type the event type to wait
    @param nticks the number of ticks to wait

*/
void
event_wait (unsigned long nticks, event_type type)
{
  char bool = 0;
  event_callout hdl;

  event_register (type, &hdl, event_notify_received, &bool);
  event_loop (nticks, &bool);
  event_unregister (&hdl);
}

