/* persist -- Persistent Storage Management
   Copyright 2001 Free Software Foundation, Inc.
   Written by Stephane Carrez (stcarrez@worldnet.fr)

This file is part of GEL.

GEL is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GEL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GEL; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include "persist.h"
#include <eeprom.h>

persistent_object* persistent_object::first;


//! Find a persistent store given its id.
//
// Returns a pointer to the persistent area or NULL if not found.
persistent_store::persistent_area_t *
persistent_store::find_store (persist_id_t id)
{
  persistent_area_t *area;

  // Walk the persistent memory from beginning to the end,
  // identify the persistent store to see if it matches 'id'.
  // If it does not match, move to the next persistent store.
  area = &start_area;
  do
    {
      // Object found.
      if (area->id == id)
	return area;

      // The invalid marker tells we are at end of the list.
      if (area->id == PERSIST_INVALID_ID
	  || area->size < sizeof (persistent_area_t))
	break;

      area = (persistent_area_t *) ((char*) area + area->size);
      area++;
    }
  while (area < &end_area);
  return 0;
}

//! Allocate a persistent area for an object.
//
// Returns a pointer to the persistent area allocated for the new object.
persistent_store::persistent_area_t *
persistent_store::allocate_store (persist_id_t id, size_t size)
{
  persistent_area_t *area;

  area = &start_area;
  do
    {
      // Found a free slot that is large enough.
      if (area->id == PERSIST_FREE_SLOT && area->size == size)
	{
	  // Save the object id.
	  eeprom_write_short (&area->id, id);
	  return area;
	}

      if (area->id == PERSIST_INVALID_ID
	  || area->size < sizeof (persistent_area_t))
	{
	  size_t space;

	  space = (size_t) ((char*) &end_area - (char*) area);
	  if (space > size + sizeof (persistent_area_t))
	    {
	      persistent_area_t *next;

	      // Save the object id and its size.
	      eeprom_write_short (&area->id, id);
	      eeprom_write_byte (&area->size, size);
	      next = (persistent_area_t *) ((char*) area + size);
	      next++;

	      // Still some room, make a new end of list.
	      if (next < &end_area && next->id != PERSIST_INVALID_ID)
		{
		  eeprom_write_short (&next->id, PERSIST_INVALID_ID);
		}
	      return area;
	    }
	  break;
	}
      area = (persistent_area_t *) ((char*) area + area->size);
      area++;
    }
  while (area < &end_area);
  return 0;
}

int
persistent_store::create (persist_id_t id, void *p, size_t size)
{
  // Truncate the size so that the size fits in a byte.
  if (size >= 256)
    size = 255;

  area = find_store (id);
  if (area != 0)
    {
      const unsigned char *src = area->data;
      unsigned char *dst = (unsigned char*) p;
      
      while (--size != 0)
	*dst++ = *src++;

      return 0;
    }
  else
    {
      area = allocate_store (id, size);
      return 1;
    }
}


void
persistent_store::save (const unsigned char *object)
{
  unsigned char *p;

  if (area == 0)
    return;

  p = area->data;
  for (unsigned char i = area->size; i != 0; i--, p++, object++)
    {
      if (*object != *p)
	{
	  eeprom_write_byte (p, *object);
	}
    }
}

void
persistent_store::revoke ()
{
  // Free the persistent store by marking the id as free.
  // Invalidate the object instance by forgetting its area.
  if (area)
    {
      eeprom_write_short (&area->id, PERSIST_FREE_SLOT);
      area = 0;
    }
}

void
persistent_object::save_all ()
{
  persistent_object *object;

  for (object = first; object; object = object->next)
    object->save (object->object);
}

