/* lsqrt-test.c -- Test for lsqrt() and lsqrt64()
   Copyright 2001 Free Software Foundation, Inc.
   Written by Stephane Carrez (stcarrez@worldnet.fr)

This file is part of GEL.

GEL is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GEL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GEL; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

/*! @page lsqrt-test Test program for the integer sqrt operations.

  @htmlonly
  Source file: <a href="lsqrt-test_c-source.html">lsqrt-test.c</a>
  @endhtmlonly

 */

#include <sys/param.h>
#include <gel/timer.h>
#include <sys/sio.h>
#include <test.h>
#include <stdio.h>
#include <benchs.h>
#include <time.h>
#include <imath.h>

#ifndef LONG_MAX
#define LONG_MAX ((1LL << 32) - 1LL)
#endif

/* Forward declarations.  */
int lsqrt_test (void);
int lsqrt64_test (void);
void lsqrt_benchs (void);


struct sqrt_val
{
  unsigned long long val;
  unsigned long long result;
};

/* Verification table.  */
struct sqrt_val sqrt_table[] =
{
  { 1, 1 },
  { 2, 1 },
  { 4, 2 },
  { 10, 3 },
  { 16, 4 },
  { 100, 10 },
  { 1000, 31 },
  { 1023, 32 },
  { 1024, 32 },
  { 10000, 100 },
  { 1000000, 1000 },
  { 1000000000, 31622 },
  { 1000000000, 31622 },
  { 10000000000, 100000 },
  { 100000000000, 316227 },
  { 1000000000000, 1000000 },
  { 10000000000000, 3162277 },
  { 100000000000000, 10000000 },
  { 1000000000000000, 31622776 },
  { 10000000000000000, 100000000 },
  { 1000000000000000000, 1000000000 }
};

#define SQRT_TABLE_SIZE ((sizeof sqrt_table) / sizeof (sqrt_table[0]))

int
lsqrt_test ()
{
  unsigned long val;
  unsigned long res;
  int i;
  
  for (i = 0; i < SQRT_TABLE_SIZE; i++)
    {
      if (sqrt_table[i].val > LONG_MAX)
        continue;

      val = sqrt_table[i].val;
      res = lsqrt (val);
      if (res != sqrt_table[i].result)
        {
          return test_error ("lsqrt(%ld) wrong got %ld",
                             val, res);
        }
    }
  return 0;
}

int
lsqrt64_test ()
{
  unsigned long long val;
  unsigned long long res;
  int i;
  
  for (i = 0; i < SQRT_TABLE_SIZE; i++)
    {
      if (sqrt_table[i].val > LONG_MAX)
        continue;

      val = sqrt_table[i].val;
      res = lsqrt64 (val);
      if (res != sqrt_table[i].result)
        {
          return test_error ("lsqrt64(%ld) wrong got %ld",
                             (long) val, (long) res);
        }
    }
  return 0;
}

void
lsqrt_benchs ()
{
  bench_t b;

  bench_init (&b);

  bench_start (&b);
  lsqrt (10000L);
  bench_stop (&b);
  bench_report (&b, "lsqrt (10000)");
  
  bench_start (&b);
  lsqrt (1000000000L);
  bench_stop (&b);
  bench_report (&b, "lsqrt (1000000000)");

  bench_start (&b);
  lsqrt64 (100000000000L);
  bench_stop (&b);
  bench_report (&b, "lsqrt64 (10000000000)");

  bench_start (&b);
  lsqrt64 (1000000000000L);
  bench_stop (&b);
  bench_report (&b, "lsqrt64 (100000000000)");
}

int
main ()
{
  int result;
  unsigned long i;

  serial_init ();

  test_depend ("serial:benchs");
  test_start ("lsqrt-test");
  test_repeat = 10;
  for (i = 0; i < test_repeat; i++)
    {
      result = lsqrt_test ();
      if (result != 0)
        break;

      result = lsqrt64_test ();
      if (result != 0)
        break;
    }
  result = test_end ("lsqrt-test");
  if (result == 0)
    {
      lsqrt_benchs ();
    }
  return result;
}
