/* triacs-test.c -- Test for triacs module
   Copyright 2001 Free Software Foundation, Inc.
   Written by Stephane Carrez (stcarrez@worldnet.fr)

This file is part of EBCS.

EBCS is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

EBCS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with EBCS; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

/*! @page lcd-test Test program for the LCD operations.

  This program tests the LCD operation by displaying patterns on
  the LCD panel.

  @htmlonly
  Source file: <a href="lcd-test_c-source.html">lcd-test.c</a>
  @endhtmlonly

 */

#include <sys/param.h>
#include <gel/timer.h>
#include <sys/sio.h>
#include <ebcs/triacs.h>
#include <test.h>
#include <stdio.h>
#include <stdlib.h>
#include <benchs.h>
#include <time.h>

int triac_blink (triac_id_t id, unsigned long time);
void triac_benchs (void);

int
triac_blink (triac_id_t id, unsigned long time)
{
  unsigned long on_count = triac_stats[id].on_counts;
  struct timeval on_time = triac_stats[id].on_total_time;
  struct timeval t;

  /* Turn the triac ON and verify it.  */
  triac_set (id, TRIAC_ON);
  if (triac_get (id) != TRIAC_ON)
    return test_error ("triac_set TRIAC_ON");

  /* Wait for `time' usec.  */
  udelay (time);

  /* Switch it OFF and verify.  */
  triac_set (id, TRIAC_OFF);
  if (triac_get (id) != TRIAC_OFF)
    return test_error ("triac_set TRIAC_OFF");

  /* Verify the statistics.  */
  if (triac_stats[id].on_counts != on_count + 1)
    return test_error ("triac_stats on_count+1");

  t.tv_usec = time % 1000000UL;
  t.tv_sec = time / 1000000UL;
  timevaladd (&on_time, &t);
  timevalsub (&on_time, &triac_stats[id].on_total_time);
  if (on_time.tv_sec == -1 && on_time.tv_usec > 0)
    {
      on_time.tv_sec = 0;
      on_time.tv_usec -= 1000000;
    }
  if (on_time.tv_sec != 0)
    return test_error ("triac_stats on_total_time seconds");

  if (on_time.tv_usec < -1000 || on_time.tv_usec > 1000)
    return test_error ("triac_stats on_total_time microseconds");

  return 0;
}

void
triac_benchs ()
{
  bench_t b;

  bench_init (&b);

  bench_start (&b);
  triac_set (TRIAC_LIGHT, TRIAC_ON);
  bench_stop (&b);
  bench_report (&b, "triac_set (TRIAC_ON)");
  
  bench_start (&b);
  triac_set (TRIAC_LIGHT, TRIAC_ON);
  bench_stop (&b);
  bench_report (&b, "triac_set (TRIAC_ON) same state");

  bench_start (&b);
  triac_set (TRIAC_LIGHT, TRIAC_OFF);
  bench_stop (&b);
  bench_report (&b, "triac_set (TRIAC_OFF)");

  bench_start (&b);
  triac_get (TRIAC_LIGHT);
  bench_stop (&b);
  bench_report (&b, "triac_get");
}

int
main ()
{
  int result;
  unsigned long i;

  serial_init ();

  timer_initialize ();
  test_repeat = 100;
  test_depend ("serial:timer:benchs");
  test_start ("triacs-test");
  for (i = 0; i < test_repeat; i++)
    {
      unsigned long dt = 10000;

      dt = rand ();
      dt = dt % 131243;
      if (triac_blink (TRIAC_LIGHT, dt) != 0)
        break;

      if (triac_blink (TRIAC_FAN, dt) != 0)
        break;
    }
  result = test_end ("triacs-test");
  if (result == 0)
    {
      triac_benchs ();
    }
  return result;
}
